// Copyright (c) Juan Carlos Daz
// All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. Neither Juan Carlos Diaz's name nor the names of his contributors
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY JUAN CARLOS DIAZ AND THE CONTRIBUTORS
// ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL JUAN CARLOS DIAZ OR THE
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "../rtbapimm-0.9/include/robot.hpp"
#include "../rtbapimm-0.9/include/consts.hpp"
#include <string>
#include <cstdlib>

const std::string robot_name = "Spi";
const int color_1 = 0xf01122;
const int color_2 = 0xfdb975;
//Estados
int rotando = 0;
int frenado = 0;
//Contador
int pared = 0;
//Cuenta las veces que aparece la pared tras perder un bot
int Cuenta = 0;
//Permite hacer Sweep tras la vuelta de reconocimiento.
int PrimeraVez = 1;

class SampleRobot : public rtb::Robot {
  struct {
    double distance;
    double last_robot_distance;
    rtb::Consts::object_type object;
    rtb::Consts::object_type last_object;
    double radar_angle;
    double last_radar_angle;
    double angulo_persecucion;
    double time;
    double speed;
    double cannon_angle;
    double energy;
    int teammate;
  } info;

//  int refract;

  void actua ();

public:
  SampleRobot (std::string name, int colour, int away_colour);
  virtual ~SampleRobot () { }
  void GameStarts ();
  void Radar (double distance, rtb::Consts::object_type object, double radar_angle);
  void Info (double time, double speed, double cannon_angle);
  void Collision (rtb::Consts::object_type object, double angle);
  void RotationReached (rtb::Consts::rotate_type what);
  void RobotInfo (double energy, int teammate);

};


/* ***************************************************
 *                      Constructor                  *
 * ************************************************* */
SampleRobot::SampleRobot (std::string name, int colour, int away_colour): Robot::Robot (name, colour, away_colour) {
}

void SampleRobot::GameStarts () {
  Accelerate (0.5);
  //si no lo tenemos delante, damos una vuelta completa
  if (info.object!=rtb::Consts::IS_ROBOT){
  	RotateAmount(rtb::Consts::ROTATE_ALL,1000,3.1416);
	rotando=1;
	//Print("ROTO!");
  }
  //Sweep (rtb::Consts::ROTATE_CANNON_RADAR, 1000, -0.5, 0.5);
  //  refract = 0;
}




/* ******************************************************************
 *     Metodo del Radar. Persigue al Robot si desaparece            *
 * **************************************************************** */
void SampleRobot::Radar (double distance, rtb::Consts::object_type object, double radar_angle) {
  info.distance = distance;
 	switch (object){
/*	  case rtb::Consts::IS_LAST_OBJECT_TYPE:Print("Objeto: IS_LAST_OBJECT_TYPE");break;
	  case rtb::Consts::IS_NOOBJECT: Print("Objeto: IS_NOOBJECT");break;
*/	  case rtb::Consts::IS_ROBOT:Print("Objeto: IS_ROBOT");
  				printf("Print Radar :: distancia: %f. angulo: %f\n", distance,radar_angle);
				break;
/*	  case rtb::Consts::IS_SHOT:Print("Objeto: IS_SHOT");break;
	  case rtb::Consts::IS_WALL:Print("Objeto: IS_WALL");break;
	  case rtb::Consts::IS_COOKIE:Print("Objeto: IS_COOKIE");break;
	  case rtb::Consts::IS_MINE:Print("Objeto: IS_MINE");break;
	  default : Print("Objeto: NI PUTA IDEA");
*/	}
	
	// Si no estamos contando las veces que vemos PARED,
	// podemos actualizar el last_object
	if (!Cuenta)  info.last_object = info.object;
	
	//Actualizacion del atributo (original de Radar)
	info.object = object;
	//Contamos las veces que detectamos PARED
	if ( (Cuenta) && (object==rtb::Consts::IS_WALL) )
		pared++;
	//Si hemos detectado 5 veces pared tras ver un robot
	//deberiamos girar para buscarlo. Angulo de giro?
	if ( pared >= 5 ){
		Print("Deberiamos dar una Vuelta de Reconocimiento");
		if (info.last_radar_angle<0){
		//	RotateAmount(rtb::Consts::ROTATE_ALL,1000,info.last_radar_angle-3);      // ??????????????? Comprobar si esto funciona
			Print("...NuevoGiro... A-3");
//			printf("Print  ----------------Escapado por la IZQ\n");
		}else{
			
		//	RotateAmount(rtb::Consts::ROTATE_ALL,1000,info.last_radar_angle+3);      // ??????????????? Comprobar si esto funciona
			Print("...NuevoGiro... A+3");
//			printf("Print  ----------------Escapado por la DER\n");
		}
		Cuenta=0;
		pared=0;
		info.last_object = info.object;
	}
	//Si perdemos al robot del foco de vision y no es por un disparo 
	//cuenta 5 veces pared seguida y da una vuelta completa para enfocar
	//de nuevo al robot.
	if (( info.last_object != info.object ) && (info.last_object==rtb::Consts::IS_ROBOT) && (info.object!=rtb::Consts::IS_SHOT) && (!Cuenta) ){
		Cuenta=1;
		Print("Robot desaparecido");
		//Vengo de disparar a kemaropa y el Sweep par.
		Sweep (rtb::Consts::ROTATE_CANNON_RADAR, 1000, -0.5, 0.5);
	}
	
	//Actualizacion del atributo (original de Radar)
	info.radar_angle = radar_angle;
	
	//Por si uso el angulo para perseguir al robot
	if (object==rtb::Consts::IS_ROBOT){
		info.last_radar_angle=radar_angle;
		info.last_robot_distance=distance;
	}
  
}

// Recibimos informacin:
void SampleRobot::Info (double time, double speed, double cannon_angle) {
  info.time = time;
  info.speed = speed;
  info.cannon_angle = cannon_angle;
//  Print("Info :: Time/Speed/CannonAngle");
  // Si el objeto es un robot:
  if (info.object != rtb::Consts::IS_ROBOT) {
    actua ();
  }
}


void SampleRobot::RobotInfo (double energy, int teammate) {
  info.energy = energy;
  info.teammate = teammate;
  //printf("Print ROBOT DETECTADO!!!  energia: %f Equipo:%d\n",energy,teammate);
  if (!info.teammate) {
    actua ();
  }
}



/****************************************************************
 *
 *          Si chocamos
 * 
*************************************************************** */
void SampleRobot::Collision (rtb::Consts::object_type object, double angle) {
  	
	// Nos ha impactado un disparo:
	if (object == rtb::Consts::IS_SHOT) {
		if (!rotando){
	      		Print ("Impactados");
    			if ((angle < -0.5) || (angle > 0.5)) { //Nos atacan por la espalda o laterales (fuera del rango de vision).
				//si nos da una ostia el robot desde lejos 
				//y fuera del angulo de vision giramos
				//if (info.last_robot_distance >= 5){
					Accelerate (1.5);
					rotando=1;
      					if (angle < -0.5){ //MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM Mirar bien !!
	      					RotateAmount(rtb::Consts::ROTATE_ALL,1000,angle);
						printf("Print Ostiazo con angulo negativo %f\n",angle);
					}
					else{
						RotateAmount(rtb::Consts::ROTATE_ALL,1000,angle);
						printf("Print Ostiazo con angulo positivo %f\n",angle);
					}
				//}//distancia
			}//angulos
		}//!rotando
    	}
	
  	// Hemos comido una Cockie volvemos a una velocidad menor
  	if (object == rtb::Consts::IS_COOKIE) {
  		Accelerate(1);
  	}

/*	switch (object){
		case rtb::Consts::IS_NOOBJECT: Print("IS_NOOBJECT");break;
		case rtb::Consts::IS_ROBOT:Print("IS_ROBOT");break;
		case rtb::Consts::IS_SHOT:Print("IS_SHOT");break;
		case rtb::Consts::IS_WALL:Print("IS_WALL");break;
		case rtb::Consts::IS_COOKIE:Print("IS_COOKIE");break;
		case rtb::Consts::IS_MINE:Print("IS_MINE");break;
		case rtb::Consts::IS_LAST_OBJECT_TYPE:Print("IS_LAST_OBJECT_TYPE");break;
		default : Print("NI PUTA IDEA");
	}
*/
	// Si nos ostiamos contra un robot calculamos 
	// por que zona fue el impacto y giramos
	if (object == rtb::Consts::IS_ROBOT) {
		printf("Print    Choque con VELOCIDAD: %f y angulo: %f\n",info.speed,angle);
		if (info.speed > 3){
			if ( (angle > 0 )  &&  ( angle < 0.5 ) )//si le damos con la derecha
				RotateAmount(rtb::Consts::ROTATE_ALL, 1000, angle+3 );
			else
				if ( (angle < 0 ) && ( angle > -0.5 ) )
					RotateAmount(rtb::Consts::ROTATE_ALL, 1000, angle-3 );
				else
					RotateAmount(rtb::Consts::ROTATE_ALL, 1000, 3.1416 );
			rotando=1;
		}
		else{//Si estamos paradillos, vamos corrigiendo el rumbo poco a poco
			if ( (angle > 0 )  &&  ( angle < 0.5 ) ){//si le damos con la derecha
				RotateAmount(rtb::Consts::ROTATE_ALL, 1000, angle+0.2 );
				rotando=1;
			}
			else
				if ( ( angle < 0 ) && ( -0.5 < angle ) ){
					RotateAmount(rtb::Consts::ROTATE_ALL, 1000, angle-0.2 );
					rotando=1;
				}
			
			
		}
	}
	
	
	//Si NoS OSTIAMOS contra una pared tenemos que girar en fundion de:
	//Si estamos rotando, lo dejamos, pq estaremos persiguiendo al robot
	//si lo que vemos es el robot, osea, el choque contra la pared es un choque lateral
	//  lo que hacemos es comprobar si la ultima vez que vimos al robot estaba cerca
	//  Si estaba cerca, giraremos a izquierda o derecha segun el ultimo valor del angulo en el que vimos al robot
	//  Si no, pues giramos 180.
	if (object == rtb::Consts::IS_WALL) {
		if (!rotando){
			//Si tenemos el robot delante, no giramos aunque choquemos
			if (info.object !=  rtb::Consts::IS_ROBOT ){
				Accelerate(0.7);
				//rotando=1;//estamos rotando
				Print("Colision -> Rotando 180");
				//si tenemos al robot cerca giramos en su sentido
//				if (info.last_robot_distance < 3){
				if (info.last_radar_angle > 0)
					RotateAmount (rtb::Consts::ROTATE_ROBOT, 1000, 3.1416 );
				else
					RotateAmount (rtb::Consts::ROTATE_ROBOT, 1000, -3.1416 );
//					info.last_robot_distance=100;
				//si no, pues damos una vuelta de reconocimiento
//				}else{
//					RotateAmount (rtb::Consts::ROTATE_ROBOT, 1000, 3.1416 );
//				}
			}
		}
	}
}

/* *****************************************************************
 * 
 *      finaliza una rotacion
 *
 *******************************************************************/
void SampleRobot::RotationReached (rtb::Consts::rotate_type what) {

	//Intento de saber que ha rotado y saber la orientacion de Sweep
/*	switch (what){
		case rtb::Consts::ROTATE_ROBOT:Print(" *** Rotado Robot");break;
		case rtb::Consts::ROTATE_CANNON:Print(" *** Rotado cannon");break;
		case rtb::Consts::ROTATE_RADAR:Print(" *** Rotado Radar");break;
		case rtb::Consts::ROTATE_ROBOT_CANNON:Print(" *** Rotado RobotYCannon");break;
		case rtb::Consts::ROTATE_CANNON_RADAR:Print(" *** Rotado CannonRadar. Sweep?");break;
		case rtb::Consts::ROTATE_ALL:Print(" *** Rotado TODO. Giro completo?");break;
		default:Print(" *** No tengo ni puta idea de que ha girado");
	}
*/	
	//Si acaba de empezar realiza barrido
	if (PrimeraVez){
		Sweep (rtb::Consts::ROTATE_CANNON_RADAR, 1000, -0.5, 0.5);
		PrimeraVez=0;
	}

//	Print("Fin Rotacion");
	if (rotando)
		rotando=0;

	if (frenado){
		frenado=0;
		Sweep (rtb::Consts::ROTATE_CANNON_RADAR, 1000, -0.5, 0.5);
//		Brake(0);
		Accelerate(0.5);
	}
}




/* ****************************************************************
 *
 *  Metodo ACTUA, dispara, esquiva, come galletas, rompe minas 
 *
 *****************************************************************/
void SampleRobot::actua () {
/*
//if (stop!=1){ //paranoya
//  if (refract)
 //   refract--;
  // */ 
  switch (info.object) {
    // Es un robot:
	case rtb::Consts::IS_ROBOT:
  //    		RotateAmount (rtb::Consts::ROTATE_ROBOT, 1000, info.radar_angle);
//      		RotateAmount (rtb::Consts::ROTATE_CANNON_RADAR, 1000, info.radar_angle);
      		RotateAmount (rtb::Consts::ROTATE_ROBOT, 1000, info.radar_angle);
		rotando=1;
		printf("Print ++++++ Persecucion con angulo: %f\n",info.radar_angle);
//		angulo_persecucion=info.radar_angle;
      		if (info.distance <= 1){
			Sweep (rtb::Consts::ROTATE_CANNON_RADAR, 1000, 0, 0);
        		Shoot (30);Shoot(30);
        		Shoot (30);Shoot(30);
        		Shoot (30);Shoot(30);
		}
      		else if (info.distance <= 2){
			Sweep (rtb::Consts::ROTATE_CANNON_RADAR, 1000, -0.0, 0.0);
        		Shoot (30);Shoot(30);
			}
      			else if (info.distance <= 3){
				Sweep (rtb::Consts::ROTATE_CANNON_RADAR, 1000, -0.1, 0.1);
        			Shoot (3);}
      				else if (info.distance <= 5){
					Sweep (rtb::Consts::ROTATE_CANNON_RADAR, 1000, -0.2, 0.2);
        				Shoot (2);}
      				else{
					Sweep (rtb::Consts::ROTATE_CANNON_RADAR, 1000, -0.2, 0.2);
        				Shoot (1);
      				}
		Print("Hemos disparado");
      		Accelerate (0.45);
    	break;
  
    
    	// Es una cookie:
    	case rtb::Consts::IS_COOKIE:
      		RotateAmount (rtb::Consts::ROTATE_ROBOT, 1000, info.radar_angle);
      		Accelerate (1);
    	break;
    
    	// Es una mina:
    	case rtb::Consts::IS_MINE:
      		Shoot (0.5);
      	break;
    
	//No activar, que dispara a sus propios disparos
	//y lia la negra pq no deja de disparar :)
   // 	case rtb::Consts::IS_SHOT:
    //  		Shoot (0.5);
      	break;
	
	// Es un muro:
    	case rtb::Consts::IS_WALL:
/*	    if (!rotando){
      		if ((info.distance <= 4) ){                      
//			Sweep (rtb::Consts::ROTATE_CANNON_RADAR, 1000, -0.0, 0.0);
			//Marcamos que estamos rotando
			rotando=1;
			//echamos los frenos
	       		Brake (1);
			Brake(0);
			Accelerate(0.1);
			frenado=1;
        		if ( (info.radar_angle < 0) ){                     
				printf("Print Angulo lateral:  %f\n",info.radar_angle);
	  			RotateAmount (rtb::Consts::ROTATE_ROBOT, 1000, (1/2)*3.1416);
        		} else {
	  			printf("Print Angulo lateral: %f\n",info.radar_angle);
	  			RotateAmount (rtb::Consts::ROTATE_ROBOT, 1000, -(1/2)*3.1416);
        		}
			
			
	
      		}//distance
	    }//if(!rotando)*/
    	break;

    	// Es un objeto desconocido:
    	case rtb::Consts::IS_NOOBJECT:
      		Accelerate (((random() % 40) - 10) / 10);
      		RotateAmount (rtb::Consts::ROTATE_ROBOT, 1000, ((random() % 40) - 20) / 10);
		rotando=1;
      		Print("Objeto Desconocido - Girando a lo loco");
	break;
  }//del switch
//}//del stop
}//del metodo


// Procedimiento main:
int main(void) {
  SampleRobot sample (robot_name, color_1, color_2);

//  sample.RobotOption( rtb::Consts::USE_NON_BLOCKING, true );
  sample.Scanner();

  return 0;
}
